/** \file generate.h
*/
#ifndef _GENERATE_H
#define _GENERATE_H

#include "objects.h"

#define HEADER_FILENAME   		"nouveau_reg.h"	///< Which file is created?
#define HEADER_LICENSE_FILENAME		"license.txt" 	///< If you want a longer license text
#define INCLUDE_IDENT_LVL1              "\t"             ///< Identation for commands
#define INCLUDE_IDENT_LVL2              "\t\t"           ///< Identation for  parameters
#define MAX_INDEX			32		///< Maximum index for define'd array
#define COLUMN_ALIGN			80		///< Alignment of values on that columns
#define TAB_SPACE			8		///< Size of a tab

/**
  \typedef nv_exclude_objects
  \brief   this struct holds the info on which objects can safely
           be skipped. This is usefull for objects like NV11, NV15 and
           NV17 which are basically the same. Without this struct we
           would dump the info 3 times, where once would be sufficient.
*/
           
typedef struct {
                 char    ex_name[100];		///< The name of the object
               } nv_exclude_objects;

/** 
 \fn	void create_header(int headerfile)
 \brief	Creates the header
 \param headerfile  filedescriptor for the HEADER_FILENAME
*/
void create_header(int headerfile);

/** 
 \fn	void create_bottom(int headerfile)
 \brief	Creates the header bottom (last few lines)
 \param headerfile  filedescriptor for the HEADER_FILENAME
*/
void create_bottom(int headerfile);

/** 
 \fn 	void create_default_include(int filedw)
 \brief add include file (if specified) into nouveau_reg.h
 \param filedw   filedescriptor for the HEADER_FILENAME
*/
void create_default_include(int filedw);

/**
 \fn	void write_text(int filed, const char *text);
 \brief	wrapper function to write texts / buffers to the file
 \param filed  the descriptor for HEADER_FILENAME
 \param text   the buffer to write into the file
*/
void write_text(int filed, const char *text);

/**
 \fn    int getArrayType(char *name)
 \brief return 0 for no Array, 1 for Array defined via A_(),
        2 for other array or 3 for ignore entry and 4 = Matrix

 We have 2 different Array types which display different names:
   V04_DX5_TEXTURED_TRIANGLE_TLVERTEX_SX(%d)   and
   NV10_TCL_PRIMITIVE_3D_TX_FORMAT(0) 
   NV10_TCL_PRIMITIVE_3D_TX_MATRIX(0)(%d) 
   
   the first one is (base + x * stride) while the other needs more
   parsing. Only index (0) will be processed, as greater indeces
   must be ignored, as they are calculated from  index(0 +x)
   TODO: What about matrices?
*/   
int isArrayType(char *name);

/**
 \fn	int get_index_offset(char * name, object_store obj);
 \brief	return the offset for the second array define
 \param name  the name of the first index "<name>(0)"
 \param obj   the obj containing all fields

 Basically look for name(1) and return its offset field
*/
int get_index_offset(char * name, object_store obj);

/**
 \fn	int get_matrix_offset(char * name, object_store obj);
 \brief	return the offset for the first dim of a matrix
 \param name  the name of the first index "<name>(0)(%d)"
 \param obj   the obj containing all fields

 Basically look for name(1)(%d) and return its offset field
*/
int get_matrix_offset(char * name, object_store obj);

/**
 \fn	void create_command_comment(int filed, object_field_store field)
 \brief write verbose command comment
*/
void create_command_comment(int filed, object_field_store field);


/**
 \fn 	void get_time(char *buffer);
 \brief enters the current time into the buffer: 24.12.2005 13:34 
*/
void get_time(char *buffer);

/**
 \fn	void write_align(int filed, int column);
 \brief	add tabs until we reach at least column
 \param filed  the descriptor for HEADER_FILENAME
 \param column which column to we want to reach at least
*/

void write_align(int filed,int column);


/**
 \fn	int check_exclude(char *obj_name);
 \brief	checks whether a given object name can be exclude from
        object generation. See \ref nv_exclude_objects
 \param obj_name  the object name to check
 \return 0 = Don't exclude, 1 = exclude 
*/
int check_exclude(char *obj_name);

/**
  \fn     void create_masks(int filed, object_field_store field);
  \brief  create the SHIFT and MASK defines as needed.
  \param  filed  the descriptor for HEADER_FILENAME
  \param  field  the field to create SHIFTS and MASKs for
  
*/
void create_masks(int filed, object_field_store field);

/** 
  \fn 	 int get_mask(int low, int high);
  \brief returns the int value for the bits requested. The range is inclusive
         example: get_mask(0, 7); will return 255
  \param low lowest bit
  \param high highest bit
*/
int get_mask(int low, int high);

/**
  \fn	 void strip_array(char *fieldname, char *result);
  \brief strip possible array indeces from the fieldname and return the result
*/
void strip_array(char *fieldname, char *result);

#endif
