/*
	Object list parser
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "nv_hw.h"

const char *engine_names[3]={
	"software object",
	"graphics object",
	"dvd object"
};

const char *dma_access[4]={
	"read/write",
	"read only",
	"write only",
	"invalid"
};

const char *dma_target[4]={
	"NV memory",
	"NV memory, tiled",
	"PCI memory",
	"AGP memory"
};

typedef struct {
	int name;
	int context;
	int data[4];
} obj_info_t;

obj_info_t *obj_infos=NULL;
int num_obj=0;

int card_family;
int tree_mode;

int get_card_family_from_file(void);
void parse_objects(void);
void parse_object(int name, int context, int *data);
void parse_software_object(int i, int data);
void parse_graphics_object(int i, int data);

int get_handle(int dma_notifier, int num_bits);

int main(int argc, char **argv)
{
	int i;
	FILE *f;
	char *file_card_obj = "card_objects.txt";

	tree_mode = 0;
	for(i=0; i<argc; i++) {
		if (strcmp("-tree", argv[i])==0) {
			tree_mode = 1;
		}
		if (strcmp("-file", argv[i])==0) {
			file_card_obj = argv[i+1];
		}
	}

	/* Retrieve card family */
	card_family = get_card_family_from_file();
	if (card_family < 0) {
		fprintf(stderr, "Can not find card family\n");
		return 1;
	}

	/* Parse object list */
	f = fopen(file_card_obj, "r");
	if (!f) {
		fprintf(stderr, "Can not open card_objects.txt\n");
		return 1;
	}

	while (!feof(f)) {
		i = num_obj++;
		obj_infos = realloc(obj_infos, sizeof(obj_info_t) * num_obj);

		fscanf(f, "%08x %08x %08x %08x %08x %08x\n",
			&obj_infos[i].name,
			&obj_infos[i].context,
			&obj_infos[i].data[0],
			&obj_infos[i].data[1],
			&obj_infos[i].data[2],
			&obj_infos[i].data[3]);
	}
	printf("%d objects loaded\n", num_obj);
	fclose(f);

	parse_objects();

	if (obj_infos) {
		free(obj_infos);
	}
	return 0;
}

int get_card_family_from_file(void)
{
	FILE *f;
	int pci_id = 0;

	/* Read pci_id from card_stdout */
	f = fopen("card_stdout.txt", "r");
	if (!f) {
		return -1;
	}	

	while(!feof(f)) {
		/* Skip comments */
		if (fgetc(f)=='#') {
			int c = 0;
			while(c!='\n') {
				c = fgetc(f);
			}
		}

		int retvalue = fscanf(f, "pci_id\t0x%08x\n", &pci_id);
		if (retvalue == 1) {
			break;
		}
	}
	fclose(f);

	return get_card_family(pci_id);
}

void parse_objects(void)
{
	int i;

	for (i=0; i<num_obj; i++) {
		parse_object(
			obj_infos[i].name,
			obj_infos[i].context,
			obj_infos[i].data
		);
	}
}

void parse_object(int name, int context, int *data)
{
	int engine=0, channel=0, dma_notify=0, i, parse_data=1;

	switch (card_family)
	{
		case NV03:
			engine = (context >> 23) & 1;
			channel = (context >> 24) & 0x1f;
			dma_notify = context & 0xffff;
			break;
		case NV04:
		case NV10:
		case NV15:
		case NV20:
		case NV30:
			engine = (context >> 16) & 0x3;
			channel = (context >> 24) & 0x1f;
			dma_notify = context & 0xffff;
			break;
		default:
			engine = (context >> 20) & 0x3;
			channel = (context >> 23) & 0x1f;
			dma_notify = context & 0xfffff;
			break;
	}

	printf("handle 0x%08x\n", name);

	printf(" context 0x%08x\n", context);
	if (context & 0x00fffff0) {
		printf("  dma notifier address: 0x%05x\n", dma_notify);
		if (!tree_mode) {
			printf("  engine: %s\n",
				(engine<3) ? engine_names[engine] : "unknown");
			printf("  channel: %d\n", channel);
		}
	} else {
		printf("  vga bios\n");
		parse_data = 0;
	}

	if (!parse_data) {
		return;
	}

	for (i=0; i<4; i++) {
		if (!tree_mode) {
			printf(" data[%d]=0x%08x\n", i,data[i]);
		}
		switch(engine) {
			case 0:
				if (!tree_mode) {
					parse_software_object(i,data[i]);
				}
				break;
			case 1:
				parse_graphics_object(i,data[i]);
				break;
			default:
				break;
		}
	}
}

void parse_software_object(int i, int data)
{
	switch(card_family) {
		case NV03:
		case NV04:
		case NV10:
		case NV15:
		case NV20:
		case NV30:
			switch(i) {
				case 0:
					printf("  class: 0x%04x\n", (data>>0) & ((1<<12)-1));
					printf("  page table: %d\n", (data>>12) & 1);
					printf("  page entry linear: %d\n", (data>>13) & 1);
					printf("  dma access: %s\n", dma_access[(data>>14) & 3]);
					printf("  dma target: %s\n", dma_target[(data>>16) & 3]);
					printf("  dma adjust: %03x\n", data>>20);
					break;
				case 1:
					printf("  dma limit: 0x%08x\n", data);
					break;
				case 2:
				case 3:
					printf("  dma page address: 0x%05x\n", data>>12);
					printf("  dma read/write: %d\n", (data>>1) & 1);
					break;
			}			
			break;
		case NV40:
		case G70:
		case G80:
			switch(i) {
				case 0:
					printf("  class: 0x%04x\n", (data>>0) & ((1<<12)-1));
					break;
				case 1:
					printf("  dma limit: 0x%08x\n", data);
					break;
				case 2:
				case 3:
					printf("  dma page address: 0x%05x\n", data>>12);
					printf("  dma read/write: %d\n", (data>>1) & 1);
					break;
			}			
			break;
	}
}

void parse_graphics_object(int i, int data)
{
	int dma_notify;

	switch(card_family) {
		case NV03:
		case NV04:
		case NV10:
		case NV15:
		case NV20:
		case NV30:
			switch(i) {
				case 0:
					printf("  class: 0x%04x\n", (data>>0) & ((1<<11)-1));
					if (!tree_mode) {
						printf("  chroma key: %d\n", (data>>12) & 1);
						printf("  user clip: %d\n", (data>>13) & 1);
						printf("  swizzle: %d\n", (data>>14) & 1);
						printf("  patch: %d\n", (data>>15) & 7);
						printf("  synchronized: %d\n", (data>>18) & 1);
						printf("  big endian: %d\n", (data>>19) & 1);
						printf("  dither: %d\n", (data>>20) & 3);
						printf("  single step: %d\n", (data>>23) & 1);
						printf("  patch valid: %d\n", (data>>24) & 1);
						printf("  surface0: %d\n", (data>>25) & 1);
						printf("  surface1: %d\n", (data>>26) & 1);
						printf("  pattern: %d\n", (data>>27) & 1);
						printf("  rop: %d\n", (data>>28) & 1);
						printf("  beta1: %d\n", (data>>29) & 1);
						printf("  beta4: %d\n", (data>>30) & 1);
					}
					break;
				case 1:
					dma_notify = (data>>16) & ((1<<16)-1);
					if (!tree_mode) {
						printf("  mono format: 0x%02x\n", (data>>0) & ((1<<8)-1));
						printf("  color format: 0x%02x\n", (data>>8) & ((1<<8)-1));
						printf("  notify address: 0x%04x\n", dma_notify);
					} else if (dma_notify) {
						printf("  uses dma notifier 0x%04x (handle 0x%08x)\n", dma_notify, get_handle(dma_notify,16));
					}
					break;
				case 2:
					dma_notify = (data>>0) & ((1<<16)-1);
					if (!tree_mode) {
						printf("  dma0 address: 0x%04x\n", dma_notify);
					} else if (dma_notify) {
						printf("  uses dma notifier 0x%04x (handle 0x%08x)\n", dma_notify, get_handle(dma_notify,16));
					}
					dma_notify = (data>>16) & ((1<<16)-1);
					if (!tree_mode) {
						printf("  dma1 address: 0x%04x\n", dma_notify);
					} else if (dma_notify) {
						printf("  uses dma notifier 0x%04x (handle 0x%08x)\n", dma_notify, get_handle(dma_notify,16));
					}
					break;
				case 3:
					break;
			}
			break;
		case NV40:
		case G70:
			switch(i) {
				case 0:
					printf("  class: 0x%04x\n", (data>>0) & ((1<<16)-1));
					printf("  user clip: %d\n", (data>>17) & 1);
					printf("  patch: %d\n", (data>>19) & 7);
					printf("  patch valid: %d\n", (data>>25) & 1);
					break;
				case 1:
				case 2:
				case 3:
					dma_notify = data & ((1<<20)-1);
					if (!tree_mode) {
						printf("  dma address: 0x%05x\n", dma_notify);
						printf("  big endian: %d\n", (data>>24) & 1);
					} else if (dma_notify) {
						printf("  uses dma notifier 0x%05x (handle 0x%08x)\n", dma_notify, get_handle(dma_notify,20));
					}
					break;
			}
			break;
		case G80:
			switch(i) {
				case 0:
					printf("  class: 0x%04x\n", (data>>0) & ((1<<16)-1));
					break;
				case 1:
				case 2:
				case 3:
					dma_notify = data & ((1<<16)-1);
					if (!tree_mode) {
						printf("  dma address: 0x%04x\n", dma_notify);
					} else if (dma_notify) {
						printf("  uses dma notifier 0x%04x (handle 0x%08x)\n", dma_notify, get_handle(dma_notify,16));
					}
					dma_notify = (data>>16) & ((1<<16)-1);
					if (!tree_mode) {
						printf("  dma address: 0x%04x\n", dma_notify);
					} else if (dma_notify) {
						printf("  uses dma notifier 0x%04x (handle 0x%08x)\n", dma_notify, get_handle(dma_notify,16));
					}
					break;
			}
			break;
	}
}

int get_handle(int dma_notifier, int num_bits)
{
	int i;

	for (i=0; i<num_obj; i++) {
		if ((obj_infos[i].context & ((1<<num_bits)-1))==dma_notifier) {
			return obj_infos[i].name;
		}
	}

	return 0;
}
