/*
	Renouveau dump parser
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>

#include "rexml.h"
#include "nv_hw.h"

typedef struct {
	int name;
	int context;
	int data[4];
	int engine;
} obj_info_t;

#define SUBCHAN_NAME_SZ 128

typedef struct {
	int hash;
	xmlNodePtr node;
	char name[SUBCHAN_NAME_SZ];
} subchan_t;

typedef struct {
	int value;
	const char *name;
} glenum_t;

enum {
	REXML_FRAGMENT_PROGRAM,
	REXML_VERTEX_PROGRAM
};

const char *filename_dump = NULL;
const char *filename_objects = "card_objects.txt";
const char *filename_stdout = "card_stdout.txt";
const char *filename_startup = "card_%04x-%04x_test_startup.txt";
const char *filename_dbxml = "renouveau.xml";
const char *filename_dasmshader = "disasm_shader";

obj_info_t *obj_infos=NULL;
int num_obj=0;

int pci_id = 0;
int card_family;
int obj_from_hash = 0;
int print_leftover = 0;
int print_as_floats = 0;
int got_objects = 0;
int num_shader_program = 0;
int cur_shader_type = -1;
subchan_t subchan[8];
char file_startup[512];
char parsed_value[512]; /* Where we store parsed dump value */

void get_pci_id(const char *filename);
int read_objects(const char *filename);
void add_object(int name, int context, int *data);
void parse_dump(const char *filename, int readsubchan);
int get_object_class(int channel);
int get_object_class_from_hash(int hash);

void rexmlParseValue(xmlNodePtr node, void *value, char *parsed_value);
void rexmlDisasmShader(int shader_type);
void rexmlParseBitfields(xmlNodePtr node, int value, char *parsed_value);
void rexmlFillEnum(xmlNodePtr node, int value, char *parsed_value);
void print_help();

int main(int argc, char **argv)
{
	int i;

        if (argc == 1)
        {
                print_help();
                exit(0);
        }

	for(i=1; i<argc; i++) {
		if (strcmp("-dump", argv[i])==0 || strcmp("--dump", argv[i])==0) {
			filename_dump = argv[++i];
                        continue;
		}
                else if (strcmp("-dbxml", argv[i])==0 || strcmp("--dbxml", argv[i])==0) {
			filename_dbxml = argv[++i];
                        continue;
		}
                else if (strcmp("-dasmshader", argv[i])==0 || strcmp("--dasmshader", argv[i])==0) {
			filename_dasmshader = argv[++i];
                        continue;
		}
                else if (strcmp("-objfromhash", argv[i])==0 || strcmp("--objfromhash", argv[i])==0) {
			obj_from_hash = 1;
                        continue;
		}
                else if (strcmp("-pciid", argv[i])==0 || strcmp("--pciid", argv[i])==0) {
			sscanf(argv[++i], "%x", &pci_id);
                        continue;
		}
                else if (strcmp("-leftover", argv[i])==0 || strcmp("--leftover", argv[i])==0) {
			print_leftover = 1;
                        continue;
		}
                else if (strcmp("-floats", argv[i])==0 || strcmp("--floats", argv[i])==0) {
			print_as_floats = 1;
                        continue;
		}
                else if (strcmp("-subchan", argv[i])==0 || strcmp("--subchan", argv[i])==0) {
			/* Will process parameters later */
			i += 2;
                        continue;
		}
		/*else if (strcmp("-objects", argv[i])==0) {
			filename_objects = argv[++i];
                        continue;
		}
                elseif (strcmp("-stdout", argv[i])==0) {
			filename_stdout = argv[++i];
                        continue;
		}*/
                else if (strcmp("-help", argv[i])==0 || strcmp("--help", argv[i])==0) {
                        print_help();
			exit(0);
		}
                else {
                        print_help();
                        exit(0);
                }
	}

	/* Retrieve card family */
	if (pci_id == 0) {
		get_pci_id(filename_stdout);
	}
	card_family = get_card_family(pci_id);
	if (card_family < 0) {
		fprintf(stderr, "Can not find card family\n");
		return 1;
	}

	memset(subchan,0,sizeof(subchan));

	if (rexmlInit(filename_dbxml)<0) {
		fprintf(stderr, "Can not load xml database %s\n", filename_dbxml);
		return 1;
	}

	got_objects = read_objects(filename_objects);

	/* Read subchannel/object from test_startup */
	snprintf(file_startup, sizeof(file_startup), filename_startup,
		(pci_id>>16) & 0xffff, pci_id & 0xffff
	);

	parse_dump(file_startup, 1);

	/* Print list of subchan hash read from test startup */
	printf("# Objects hash in subchan after test_startup\n");
	for (i=0; i<8; i++) {
		if (!subchan[i].name[0]) {
			snprintf(subchan[i].name, SUBCHAN_NAME_SZ,
				"NvObject%04x",
				(int)(get_object_class(i) & 0x7fffffffUL));
		}
		printf("#  Subchan[%d].hash = 0x%08x: %s\n", i, subchan[i].hash, subchan[i].name);
	}

	{
		int new_hash = 0;
		for(i=0; i<argc; i++) {
			if (strcmp("-subchan", argv[i])==0 || strcmp("--subchan", argv[i])==0) {
				int n_hash;
				sscanf(argv[i+2], "%x", &n_hash);
				subchan[atoi(argv[i+1])].hash = n_hash << 8;
				memset(subchan[atoi(argv[i+1])].name, 0, SUBCHAN_NAME_SZ);
				new_hash = 1;
			}
		}
		if (new_hash) {
			printf("# Objects hash in subchan after changes\n");
			for (i=0; i<8; i++) {
				if (!subchan[i].name[0]) {
					snprintf(subchan[i].name,
						SUBCHAN_NAME_SZ,
						"NvObject%04x",
						(int)(get_object_class(i) &
							0x7fffffffUL));
				}
				printf("#  Subchan[%d].hash = 0x%08x: %s\n", i, subchan[i].hash, subchan[i].name);
			}
		}
	}

	if (filename_dump) {
		parse_dump(filename_dump, 0);
	}

	if (obj_infos) {
		free(obj_infos);
	}

	rexmlShutdown();
	return 0;
}

void print_help()
{
        printf("Requires:\n");
        printf(" card_objects.txt, card_stdout.txt\n");
        printf("Optional:\n");
        printf(" card-PCIID-test_startup.txt for your card\n");
        printf(" in current directory (if objfromhash not used)\n");
        printf("Parameters (accepts both - and --):\n");
        printf(" -dbxml <filename.xml> (xml database for parser)\n");
        printf(" -dump <filename-dump.txt> (dump to parse)\n");
        printf(" -dasmshader <path/to/disasm_shader> (shader disassembler)\n");
        printf(" -objfromhash (use object type from hash value)\n");
        /* hashobjtype = XX in a yyyyXXyy hash */
        printf(" -pciid <n> (force pci id, must be 10deXXXX)\n");
        printf(" -subchan <n> <hashobjtype> (force hash, use with objfromhash)\n");
        printf(" -leftover (print remaining unknown bits)\n");
        printf(" -floats (print values as floats)\n");
}

void get_pci_id(const char *filename)
{
	FILE *f;

	/* Read pci_id from card_stdout */
	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return;
	}	

	while(!feof(f)) {
		/* Skip comments */
		if (fgetc(f)=='#') {
			int c = 0;
			while(c!='\n') {
				c = fgetc(f);
			}
		}

		int retvalue = fscanf(f, "pci_id\t0x%08x\n", &pci_id);
		if (retvalue == 1) {
			break;
		}
	}
	fclose(f);
}

int read_objects(const char *filename)
{
	FILE *f;

	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return 0;
	}

	while (!feof(f)) {
		int name, context, data[4];

		fscanf(f, "%08x %08x %08x %08x %08x %08x\n",
			&name, &context, &data[0], &data[1], &data[2], &data[3]);

		add_object(name, context, data);
	}
	fclose(f);

	return 1;
}

void add_object(int name, int context, int *data)
{
	int i = num_obj++;
	obj_infos = realloc(obj_infos, sizeof(obj_info_t) * num_obj);

	obj_infos[i].name = name;
	obj_infos[i].context = context;
	obj_infos[i].data[0] = data[0];
	obj_infos[i].data[1] = data[1];
	obj_infos[i].data[2] = data[2];
	obj_infos[i].data[3] = data[3];

	switch (card_family)
	{
		case NV03:
			obj_infos[i].engine = (obj_infos[i].context >> 23) & 1;
			break;
		case NV04:
		case NV10:
		case NV15:
		case NV20:
		case NV30:
			obj_infos[i].engine = (obj_infos[i].context >> 16) & 0x3;
			break;
		default:
			obj_infos[i].engine = (obj_infos[i].context >> 20) & 0x3;
			break;
	}
}

void printStringDummy(const char *value)
{
}

void printHeaderDummy(int size, int channel, int offset, int increment, int value)
{
}

void printParamDummy(int channel, int offset, int value)
{
	if (offset==0) {
		int object_id;

		subchan[channel].hash = value;
		object_id = get_object_class(channel);
		subchan[channel].node = rexmlGetObjectNode(object_id & 0x7fffffff);
		if (subchan[channel].node) {
			xmlChar *objname = xmlGetProp(
				subchan[channel].node,
				BAD_CAST "name");
			strncpy(subchan[channel].name,
				(char*)objname,
				SUBCHAN_NAME_SZ);
			subchan[channel].name[SUBCHAN_NAME_SZ-1] = '\0';
			xmlFree(objname);
		} else {
			snprintf(subchan[channel].name,
				SUBCHAN_NAME_SZ,
				"NvObject%04x",
				object_id & 0x7fffffff);
		}
	}
}

void printStringParser(const char *value)
{
	printf("%s\n", value);
}

void printHeaderParser(int size, int channel, int offset, int increment, int value)
{
	printf("%08x  size %d, subchannel %d (0x%08x),offset 0x%04x,%s\n",
		value, size,
		channel, offset ? subchan[channel].hash : 0,
		offset, increment ? "increment" : "constant");
}

void printParamParser(int channel, int offset, int value)
{
	printParamDummy(channel, offset, value);

	printf("%08x    ", value);

	if (offset==0) {
		int object_id = get_object_class(channel);

		printf("%s mapped to subchannel %d %s\n", subchan[channel].name,
			channel, (object_id & (1<<31)) ? "(GUESS)" : "");
	} else {
		int is_array = 0, stride=4;
		int start_offset = offset;
		xmlNodePtr func_node = NULL;
		xmlChar *func_name = NULL;

		/* Object name on this subchannel */
		printf("%s", subchan[channel].name);
					
		func_node = rexmlGetFunctionNode(subchan[channel].node, offset,
			&is_array, &start_offset, &stride);
		if (func_node) {
			func_name = xmlGetProp(func_node, BAD_CAST "name");
		}
		if (func_name) {
			/* Function name inside object if known */
			printf(".%s", func_name);
			xmlFree(func_name);
		} else {
			/* Or offset if unknown */
			printf("[0x%04x/4]", offset);
		}

		/* If array, print index */
		if (is_array) {
			int array_index = (offset-start_offset)/stride;
			printf("[%d]", array_index);
		}

		/* Then parsed value from type */
		memset(parsed_value,0,sizeof(parsed_value));
		if (func_node) {
			rexmlParseValue(func_node, &value, parsed_value);
		}

		if (parsed_value[0]) {
			printf(" = %s", parsed_value);
		}

		/* Print value as float ? */
		if (print_as_floats) {
			printf(" (%f)", *((float *) &value));
		}

		printf("\n");

		if (cur_shader_type!=-1) {
			rexmlDisasmShader(cur_shader_type);
			++num_shader_program;
			cur_shader_type = -1;
		}
	}
}

#define FIFO_INST_PATH "/renouveau/fifo/instruction"

void parse_dump(const char *filename, int readsubchan)
{
	rexmlParseFuncs_t parseFuncs;

	if (readsubchan) {
		parseFuncs.printString = printStringDummy;
		parseFuncs.printHeader = printHeaderDummy;
		parseFuncs.printParam = printParamDummy;
	} else {
		parseFuncs.printString = printStringParser;
		parseFuncs.printHeader = printHeaderParser;
		parseFuncs.printParam = printParamParser;
	}

	rexmlParse(filename, &parseFuncs);
}

int get_object_class(int channel)
{
	if (obj_from_hash) {
		return ((subchan[channel].hash >> 8) & 0xff) | 0x80000000UL;
	}

	return get_object_class_from_hash(subchan[channel].hash);
}

int get_object_class_from_hash(int hash)
{
	/* Search this hash in object list */
	int i, ctx, datas[4] = {0,0,0,0};

	for (i=0; i<num_obj; i++) {
		if (obj_infos[i].name != hash) {
			continue;
		}

		int class_mask = ((1<<12)-1);

		if (obj_infos[i].engine == 1) {
			/* graphics, nv40/g80 has 16bits */
			switch(card_family) {
				case NV40:
				case G70:
				case G80:
					class_mask = ((1<<16)-1);
					break;
				default:
					break;
			}
		}

		return obj_infos[i].data[0] & class_mask;
	}

	/* Create a dummy object from hash */
	datas[0] = (hash>>8) & 0xff;
	ctx = 0x82000000 + datas[0];
	add_object(hash, ctx, datas);

	/* Return object from hash if not found */
	return ((hash >> 8) & 0xff) | 0x80000000UL;
}

/*
	Parse node value
	FIXME: reserved space for *parsed_value?
*/

void rexmlParseValue(xmlNodePtr node, void *value, char *parsed_value)
{
	xmlChar *attValue;

	attValue = xmlGetProp(node, BAD_CAST "type");
	if (!attValue) {
		return;
	}

	if (xmlStrcmp(BAD_CAST "boolean", attValue)==0) {
		/* False/true value */
		sprintf(parsed_value,"%s", *((int *)value) ? "TRUE" : "FALSE");
	} else if (xmlStrcmp(BAD_CAST "float", attValue)==0) {
		/* Float value */
		sprintf(parsed_value,"%f", *((float *)value));
	} else if (xmlStrcmp(BAD_CAST "enum", attValue)==0) {
		rexmlFillEnum(node, *((int *)value), parsed_value);
	} else if (xmlStrcmp(BAD_CAST "hexa", attValue)==0) {
		/* Int hexa value */
		sprintf(parsed_value,"0x%08x", *((int *)value));
	} else if (xmlStrcmp(BAD_CAST "int", attValue)==0) {
		/* Int value */
		sprintf(parsed_value,"%d", *((int *)value));
	} else if (xmlStrcmp(BAD_CAST "object", attValue)==0) {
		/* Object type from hash */		
		int object_id = get_object_class_from_hash(*((int *)value)) & 0x7fffffff;
		xmlNodePtr node = rexmlGetObjectNode(object_id);

		if (node) {
			xmlChar *objname = xmlGetProp(node, BAD_CAST "name");
			/* FIXME: buffer overflow in sprintf() */
			sprintf(parsed_value, "%s", objname);
			xmlFree(objname);
		} else {
			/* FIXME: buffer overflow in sprintf() */
			sprintf(parsed_value, "Unknown object 0x%04x", object_id);
		}
	} else if (xmlStrcmp(BAD_CAST "bitfield", attValue)==0) {
		/* This node contains bitfields */
		rexmlParseBitfields(node, *((int *)value), parsed_value);
	} else if (xmlStrcmp(BAD_CAST "program", attValue)==0) {
		sprintf(parsed_value,"0x%08x", *((int *)value));

		/* Dump shader */
		/*xmlChar *progTypeValue = xmlGetProp(node, BAD_CAST "program_type");
		if (progTypeValue) {
			xmlFree(progTypeValue);
		}*/

		/*cur_shader_type = REXML_FRAGMENT_PROGRAM;
		cur_shader_type = REXML_VERTEX_PROGRAM;*/
	}

	xmlFree(attValue);
}

void rexmlDisasmShader(int shader_type)
{
	const char *shader_family = NULL;
	char *filename_shader, *filename_dump1;

	switch(card_family) {
		case NV20:
			shader_family = "NV20";
			break;
		case NV30:
			shader_family = "NV30";
			break;
		case NV40:
			shader_family = "NV40";
			break;
		case G70:
			shader_family = "G70";
			break;
		default:
			fprintf(stderr, "Unsupported shader program\n");
			return;
	}

	pid_t pid = fork();
	if (pid>0) {
		/* Wait child to terminate */
		waitpid(-1, NULL, WNOHANG);
	} else if (pid==0) {
		filename_dump1 = (char *)calloc(strlen(filename_dump), sizeof(char));
		if (!filename_dump1) {
			fprintf(stderr, "Can not allocate mem for shader dump\n");
			exit(0);
		}

		strncpy(filename_dump1, filename_dump, strlen(filename_dump)-4);

		filename_shader = (char *)calloc(strlen(filename_dump)+16, sizeof(char));
		if (!filename_shader) {
			fprintf(stderr, "Can not allocate mem for shader dump\n");
			free(filename_dump1);
			exit(0);
		}

		/* dump name is card_10de-xxxx-test_name.txt
			we generate card_10de-xxxx-test_name_prg<n>.txt
		*/
		sprintf(filename_shader, "%s_prg%d.txt", filename_dump1, num_shader_program);

		execlp(
			filename_dasmshader,
			filename_dasmshader,
			shader_family, /* family */
			(shader_type==REXML_FRAGMENT_PROGRAM) ? "fragment" : "vertex", /* type */
			filename_shader, /* file */
			NULL
		);

		free(filename_dump1);
		free(filename_shader);
		exit(0);
	} else {
		fprintf(stderr, "Can not fork\n");
	}
}

/*
	Parse bitfields of a node
*/

void rexmlParseBitfields(xmlNodePtr node, int value, char *parsed_value)
{
	xmlNodePtr arrsz = node->children;
	xmlChar *attValue, *fieldname, *fieldtype;
	char tmp[32];
	int highbit = -1, lowbit = -1, field, mask, firsttime = 1;
	uint32_t bitfield_mask_used = 0;

 	for(; arrsz; arrsz=arrsz->next) {
		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "bitfield")!=0) {
			continue;
		}

		fieldname = xmlGetProp(arrsz, BAD_CAST "name");
		if (!fieldname) {
			continue;
		}

		fieldtype = xmlGetProp(arrsz, BAD_CAST "type");

		attValue = xmlGetProp(arrsz, BAD_CAST "high");
		if (attValue) {
			highbit = atoi((char*)attValue);
			xmlFree(attValue);
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "low");
		if (attValue) {
			lowbit = atoi((char*)attValue);
			xmlFree(attValue);
		}

		if ((lowbit<0) || (highbit<0)) {
			continue;
		}

		mask = (highbit==31) ? -1 : (1<<(highbit+1))-1;
		field = ((unsigned int) (value & mask)) >> lowbit;
		bitfield_mask_used |= ((unsigned int) mask) & (0xffffffffUL << lowbit);

		/* Add separator if needed */
		if (!firsttime) {
			strcat(parsed_value, " | ");
		}

		strcat(parsed_value, fieldname);
		strcat(parsed_value, "=");

		/* Default to hexa */
		sprintf(tmp, "0x%08x", field);
		if (xmlStrcmp(fieldtype, BAD_CAST "int")==0) {
			sprintf(tmp, "%d", field);
		} else if (xmlStrcmp(fieldtype, BAD_CAST "boolean")==0) {
			sprintf(tmp, "%s", field ? "TRUE" : "FALSE");
		} else if (xmlStrcmp(fieldtype, BAD_CAST "program")==0) {
			sprintf(tmp, "0x%08x", field<<lowbit);
		} else if (xmlStrcmp(fieldtype, BAD_CAST "enum")==0) {
			rexmlFillEnum(arrsz, field, tmp);
		}
		strcat(parsed_value, tmp);

		if (fieldtype) {
			xmlFree(fieldtype);
		}
		xmlFree(fieldname);
		firsttime = 0;
	}

	/* Now print bitfield left */
	if (print_leftover) {
		value &= ~bitfield_mask_used;
		if (bitfield_mask_used != 0xffffffffUL) {
			if (!firsttime) {
				strcat(parsed_value, " | ");
			}
			strcat(parsed_value, "leftover=");
			sprintf(tmp, "0x%08x/0x%08x", value, bitfield_mask_used);
			strcat(parsed_value, tmp);
		}
	}
}

/*
	Fill value from enumeration
*/

void rexmlFillEnum(xmlNodePtr node, int value, char *parsed_value)
{
	xmlXPathObjectPtr pathobj;
	xmlNodePtr list, arrsz;
	xmlChar *attValue;

	xmlChar *enum_name = xmlGetProp(node, BAD_CAST "enum_name");
	if (!enum_name) {
		return;
	}

	pathobj = rexmlSearchNode(BAD_CAST "/renouveau/enums");
	if (!pathobj) {
		xmlFree(enum_name);
		return;
	}

	/* Search an enumeration list with this name */
	list = rexmlSearchNodeAttrib(pathobj->nodesetval->nodeTab[0],
		"list", "name", (char*)enum_name);

	if (!list) {
		xmlXPathFreeObject(pathobj);
		xmlFree(enum_name);
		return;
	}

	/* Now search an item with this value */
	arrsz = list->children;
 	for(; arrsz; arrsz=arrsz->next) {
		int found = 0, item_value = 0;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "item")!=0) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "value");
		if (attValue) {
			if (sscanf((char*)attValue, "0x%x", &item_value)) {
				found = (item_value == value);
			}
			xmlFree(attValue);
		}

		if (found) {
			attValue = xmlGetProp(arrsz, BAD_CAST "name");
			if (attValue) {
				sprintf(parsed_value, "%s", attValue);
				xmlFree(attValue);
			}
			break;
		}
	}

	xmlXPathFreeObject(pathobj);
	xmlFree(enum_name);
}
