/*
	renouveau.xml related functions
*/

#include <string.h>

#include "rexml.h"

/*--- Variables ---*/

static xmlDocPtr redbxml = NULL;

/*--- Functions ---*/

int rexmlInit(const char *filename)
{
	LIBXML_TEST_VERSION
	xmlInitParser();

	redbxml = xmlParseFile(filename);
	if (redbxml==NULL) {
		return -1;
	}

	return 0;
}

void rexmlShutdown(void)
{
	if (redbxml!=NULL) {
		xmlFreeDoc(redbxml);
		redbxml=NULL;
	}

	xmlCleanupParser();
	xmlMemoryDump();
}

/* Search a path */

xmlXPathObjectPtr rexmlSearchNode(xmlChar *xpath)
{
	xmlXPathContextPtr context;
	xmlXPathObjectPtr pathobj;

	context = xmlXPathNewContext(redbxml);
	pathobj = xmlXPathEvalExpression(xpath, context);
	if (!pathobj) {
		return NULL;
	}
	xmlXPathFreeContext(context);

	return pathobj;
}

/*
	Parse a dump
*/

void rexmlParse(const char *filename, rexmlParseFuncs_t *rexmlParseFuncs)
{
	FILE *f;
	xmlXPathObjectPtr fifo_inst_path = NULL;
	xmlNodePtr fifo_inst_node = NULL;
	xmlNodeSetPtr nodeset;
	int num_nops = 0;
	int latest_inst = 0;
	int skip_values = 0;
	char tmp_str[512];

	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Can not open %s\n", filename);
		return;
	}
	
	fifo_inst_path = rexmlSearchNode(BAD_CAST "/renouveau/fifo/instruction");
	if (!fifo_inst_path) {
		fprintf(stderr, "Can not find xml path to fifo instruction\n");
		fclose(f);
		return;
	}

	nodeset = fifo_inst_path->nodesetval;
	if (!xmlXPathNodeSetIsEmpty(nodeset)) {
		if (nodeset->nodeNr>=1) {
			fifo_inst_node = nodeset->nodeTab[0];
		}
	}

	if (!fifo_inst_node) {
		fprintf(stderr, "Can not find node for fifo instruction\n");
		xmlXPathFreeObject(fifo_inst_path);
		fclose(f);
		return;
	}

	while (!feof(f)) {
		unsigned int dump_value;
		int is_jump, i;
		int size, channel, offset, increment;

		/* Skip comments */
		if (fgetc(f)=='#') {
			int c = '#';
			unsigned int pos = 0;

			memset(tmp_str, 0, sizeof(tmp_str));
			while (c != '\n') {
				if (pos<sizeof(tmp_str)-1) {
					tmp_str[pos++] = c;
				}
				c = fgetc(f);
			}
			/* tmp_str has no carriage return */
			rexmlParseFuncs->printString(tmp_str);
		}

		if (fscanf(f, "%08x", &dump_value) != 1) {
			continue;
		}

		if ((dump_value == 0) && (skip_values==0)) {
			num_nops++;
			continue;
		}

		if (num_nops && (skip_values==0)) {
			snprintf(tmp_str, sizeof(tmp_str),
				"# %d NOPs", num_nops);
			rexmlParseFuncs->printString(tmp_str);
			num_nops = 0;
		}

		is_jump = rexmlGetBitfield(fifo_inst_node, "jump", dump_value);

		if (is_jump) {
			int new_pos = rexmlGetBitfield(fifo_inst_node, "jump_offset", dump_value)<<2;
#if 1
			snprintf(tmp_str, sizeof(tmp_str),
				"%08x  # Ignore jump to position 0x%08x",
				dump_value, new_pos);
			rexmlParseFuncs->printString(tmp_str);
			continue;
#else
			if (new_pos) {
				snprintf(tmp_str, sizeof(tmp_str),
					"# Jump to fifo position 0x%08x",
					new_pos);
				rexmlParseFuncs->printString(tmp_str);
			} else {
				snprintf(tmp_str, sizeof(tmp_str),
					"# Replay latest instruction");
				rexmlParseFuncs->printString(tmp_str);
				dump_value = latest_inst;
			}
#endif
		}

		latest_inst = dump_value;

		size = rexmlGetBitfield(fifo_inst_node, "size", dump_value);
		channel = rexmlGetBitfield(fifo_inst_node, "subchannel", dump_value);
		offset = rexmlGetBitfield(fifo_inst_node, "index", dump_value)<<2;
		increment = (rexmlGetBitfield(fifo_inst_node, "method", dump_value)==0);

		/* Skip stuff nvidia send to trigger a software method */
		if ((size==0) && (skip_values==0)) {
			switch(dump_value) {
				/*	0xbee4 + 8 values on nvxx_image_blit channel */
				case 0xbee4:
					skip_values=9;
					break;
				/*	0xc1e4 + 4 values on nvxx_context_surface_2d channel */
				case 0xc1e4:
				/*	0xbee0 + 4 values on nvxx_beta_solid channel */
				case 0xbee0:
					skip_values=5;
					break;
				/*	0x3740 on nvxx_tcl channel */
				case 0x3740:
					skip_values=1;
					break;
			}
		}
		if (skip_values>0) {
			snprintf(tmp_str, sizeof(tmp_str),
				"%08x  skip", dump_value);
			rexmlParseFuncs->printString(tmp_str);
			skip_values--;
			continue;
		}

		rexmlParseFuncs->printHeader(size, channel, offset, increment, dump_value);

		for (i=0; i<size; i++) {
			if (feof(f)) {
				break;
			}

			/* Skip comments */
			if (fgetc(f)=='#') {
				int c = '#';
				unsigned int pos = 0;

				memset(tmp_str, 0, sizeof(tmp_str));
				while (c != '\n') {
					if (pos<sizeof(tmp_str)-1) {
						tmp_str[pos++] = c;
					}
					c = fgetc(f);
				}
				/* tmp_str has no carriage return */
				rexmlParseFuncs->printString(tmp_str);
			}

			if (fscanf(f, "%08x", &dump_value) != 1) {
				--i;	/* To deal with empty lines */
				continue;
			}

			rexmlParseFuncs->printParam(channel, offset, dump_value);

			if (increment) {
				offset += 4;
			}
		}
	}
	fclose(f);

	xmlXPathFreeObject(fifo_inst_path);
}

/*
	Search a node bitfield in given 'node' with 'attr_name' name attribute
		Return 0 if error, or parsed value from bitfield inside value
*/

int rexmlGetBitfield(xmlNodePtr node, const char *attr_name, int value)
{
	xmlNodePtr node_attrib;
	xmlChar *attValue;
	int retval = 0;
	int highbit = -1, lowbit = -1, mask;

	node_attrib = rexmlSearchNodeAttrib(node, "bitfield", "name", attr_name);

	if (node_attrib==NULL) {
		return retval;
	}

	attValue = xmlGetProp(node_attrib, BAD_CAST "high");
	if (attValue) {
		highbit = atoi((char*)attValue);
		xmlFree(attValue);
	}

	attValue = xmlGetProp(node_attrib, BAD_CAST "low");
	if (attValue) {
		lowbit = atoi((char*)attValue);
		xmlFree(attValue);
	}

	if ((lowbit<0) || (highbit<0)) {
		return retval;
	}

	mask = ((highbit==31) ? -1 : (1<<(highbit+1))-1);
	retval = ((unsigned int) (value & mask)) >> lowbit;

	return retval;
}

/*
	Search a child node with given node name and attribute
*/

xmlNodePtr rexmlSearchNodeAttrib(xmlNodePtr parent, const char *node_name,
	const char *attr_name, const char *attr_value)
{
	xmlNode *arrsz = parent->children;

	for(; arrsz; arrsz=arrsz->next) {
		xmlChar *attValue;
		int found = 0;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST node_name)!=0) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST attr_name);
		if (attValue) {
			found = (xmlStrcmp(attValue, BAD_CAST attr_value)==0);
			xmlFree(attValue);
		}

		if (found) {
			return arrsz;
		}
        }
	
	return NULL;
}

/*
	Return function node of an object given index
*/

xmlNodePtr rexmlGetFunctionNode(xmlNodePtr parent, int index,
	int *is_array, int *start_offset, int *stride)
{
	xmlNodePtr node = NULL, arrsz;
	xmlChar *attValue;
	int parent_id, found = 0, offset, size;

 	if (!parent) {
		return NULL;
	}

	arrsz = parent->children;
 	*is_array = 0;
 
 	for(; arrsz; arrsz=arrsz->next) {
		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "reg32")!=0) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "size");
		size = 1;
		if (attValue) {
			if (!sscanf((char*)attValue, "%d", &size)) {
				size = 1;
			}
			xmlFree(attValue);
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "stride");
		*stride = 4;
		if (attValue) {
			if (!sscanf((char*)attValue, "%d", stride)) {
				*stride = 4;
			}
			xmlFree(attValue);
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "offset");
		if (attValue) {
			if (sscanf((char*)attValue, "0x%x", &offset)) {
				int i;
				for (i=0; i<size; i++) {
					if (index == offset+i* (*stride)) {
						found = 1;
						break;
					}
				}
			}
			xmlFree(attValue);
		}

		if (found) {
			*is_array = (size>1);
			if (is_array) {
				*start_offset = offset;
			}
			node = arrsz;
			break;
		}
	}

	if (!node) {
		/* Try parent objects now */
		attValue = xmlGetProp(parent, BAD_CAST "parent");
		if (attValue) {
			if (sscanf((char*)attValue, "0x%x", &parent_id)) {
				node = rexmlGetFunctionNode(rexmlGetObjectNode(parent_id), index, is_array, start_offset, stride);
			}
			xmlFree(attValue);
		}
	}

	return node;
}

/*
	Return object node from object_id
*/

xmlNodePtr rexmlGetObjectNode(int object_id)
{
	xmlXPathObjectPtr pathobj;
	xmlNodePtr node = NULL, arrsz;

	pathobj = rexmlSearchNode(BAD_CAST "/renouveau");
	if (!pathobj) {
		return node;
	}

	arrsz = pathobj->nodesetval->nodeTab[0]->children;

	for(; arrsz; arrsz=arrsz->next) {
		xmlChar *attValue;
		int found = 0, id;

		if ( (arrsz->type != XML_ELEMENT_NODE) || !arrsz->name) {
			continue;
		}

		if (xmlStrcmp(arrsz->name, BAD_CAST "object")!=0) {
			continue;
		}

		attValue = xmlGetProp(arrsz, BAD_CAST "id");
		if (sscanf((char*)attValue, "0x%x", &id)) {
			found = (id == object_id);
		}
		xmlFree(attValue);

		if (found) {
			node = arrsz;
			break;
		}
	}

	xmlXPathFreeObject(pathobj);

	return node;
}
